@@ -1,8 +1,8 @@
+# SPDX-License-Identifier: GPL-2.0-only
 # This file is part of Scapy
-# See http://www.secdev.org/projects/scapy for more information
+# See https://scapy.net/ for more information
 # Copyright (C) Andreas Korb <andreas.d.korb@gmail.com>
 # Copyright (C) Nils Weiss <nils@we155.de>
-# This program is published under a GPLv2 license
 
 # scapy.contrib.status = skip
 
@@ -30,7 +30,7 @@ class OBD_S09_PR(Packet):
     ]
 
     def answers(self, other):
-        return other.__class__ == OBD_S09 \
+        return isinstance(other, OBD_S09) \
             and all(r.iid in other.iid for r in self.data_records)
 
 
