@@ -1,7 +1,7 @@
+# SPDX-License-Identifier: GPL-2.0-only
 # This file is part of Scapy
-# See http://www.secdev.org/projects/scapy for more information
+# See https://scapy.net/ for more information
 # Copyright (C) Lucas Preston <lucas.preston@infinite.io>
-# This program is published under a GPLv2 license
 
 # scapy.contrib.description = Network File System (NFS) v3
 # scapy.contrib.status = loads
@@ -12,7 +12,6 @@ from scapy.packet import Packet, bind_layers
 from scapy.fields import IntField, IntEnumField, FieldListField, LongField, \
     XIntField, XLongField, ConditionalField, PacketListField, StrLenField, \
     PacketField
-from scapy.modules.six import integer_types
 
 nfsstat3 = {
     0: 'NFS3_OK',
@@ -58,7 +57,7 @@ ftype3 = {
 
 
 def loct(x):
-    if isinstance(x, integer_types):
+    if isinstance(x, int):
         return oct(x)
     if isinstance(x, tuple):
         return "(%s)" % ", ".join(map(loct, x))
@@ -442,7 +441,7 @@ class READDIRPLUS_Reply(Packet):
         ),
         ConditionalField(
             PacketListField(
-                'files', None, cls=File_From_Dir_Plus,
+                'files', None, File_From_Dir_Plus,
                 next_cls_cb=lambda pkt, lst, cur, remain:
                 File_From_Dir_Plus if pkt.value_follows == 1 and
                 (len(lst) == 0 or cur.value_follows == 1) and
@@ -722,7 +721,7 @@ class READDIR_Reply(Packet):
         ),
         ConditionalField(
             PacketListField(
-                'files', None, cls=File_From_Dir,
+                'files', None, File_From_Dir,
                 next_cls_cb=lambda pkt, lst, cur, remain:
                 File_From_Dir if pkt.value_follows == 1 and
                 (len(lst) == 0 or cur.value_follows == 1) and
