@@ -1,24 +1,37 @@
+# SPDX-License-Identifier: GPL-2.0-only
 # This file is part of Scapy
-# See http://www.secdev.org/projects/scapy for more information
+# See https://scapy.net/ for more information
 # Copyright (C) Philippe Biondi <phil@secdev.org>
-# This program is published under a GPLv2 license
 
 """
 Direct Access dictionary.
 """
 
-from __future__ import absolute_import
-from __future__ import print_function
 from scapy.error import Scapy_Exception
-import scapy.modules.six as six
 from scapy.compat import plain_str
 
+# Typing
+from typing import (
+    Any,
+    Dict,
+    Generic,
+    Iterator,
+    List,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+)
+from scapy.compat import Self
+
+
 ###############################
 #  Direct Access dictionary   #
 ###############################
 
 
 def fixname(x):
+    # type: (Union[bytes, str]) -> str
     """
     Modifies a string to make sure it can be used as an attribute name.
     """
@@ -38,7 +51,11 @@ class DADict_Exception(Scapy_Exception):
     pass
 
 
-class DADict(object):
+_K = TypeVar('_K')  # Key type
+_V = TypeVar('_V')  # Value type
+
+
+class DADict(Generic[_K, _V]):
     """
     Direct Access Dictionary
 
@@ -54,65 +71,96 @@ class DADict(object):
 
         ETHER_TYPES.IPv4 -> 2048
     """
+    __slots__ = ["_name", "d"]
+
     def __init__(self, _name="DADict", **kargs):
+        # type: (str, **Any) -> None
         self._name = _name
-        self.update(kargs)
+        self.d = {}  # type: Dict[_K, _V]
+        self.update(kargs)  # type: ignore
 
     def ident(self, v):
+        # type: (_V) -> str
         """
         Return value that is used as key for the direct access
         """
-        return fixname(v)
+        if isinstance(v, (str, bytes)):
+            return fixname(v)
+        return "unknown"
 
     def update(self, *args, **kwargs):
-        for k, v in six.iteritems(dict(*args, **kwargs)):
-            self[k] = v
+        # type: (*Dict[_K, _V], **Dict[_K, _V]) -> None
+        for k, v in dict(*args, **kwargs).items():
+            self[k] = v  # type: ignore
 
     def iterkeys(self):
-        for x in six.iterkeys(self.__dict__):
+        # type: () -> Iterator[_K]
+        for x in self.d:
             if not isinstance(x, str) or x[0] != "_":
                 yield x
 
     def keys(self):
+        # type: () -> List[_K]
         return list(self.iterkeys())
 
     def __iter__(self):
+        # type: () -> Iterator[_K]
         return self.iterkeys()
 
     def itervalues(self):
-        return six.itervalues(self.__dict__)
+        # type: () -> Iterator[_V]
+        return self.d.values()  # type: ignore
 
     def values(self):
+        # type: () -> List[_V]
         return list(self.itervalues())
 
     def _show(self):
+        # type: () -> None
         for k in self.iterkeys():
             print("%10s = %r" % (k, self[k]))
 
     def __repr__(self):
+        # type: () -> str
         return "<%s - %s elements>" % (self._name, len(self))
 
     def __getitem__(self, attr):
-        return self.__dict__[attr]
+        # type: (_K) -> _V
+        return self.d[attr]
 
     def __setitem__(self, attr, val):
-        self.__dict__[attr] = val
+        # type: (_K, _V) -> None
+        self.d[attr] = val
 
     def __len__(self):
-        return len(self.__dict__)
+        # type: () -> int
+        return len(self.d)
 
     def __nonzero__(self):
+        # type: () -> bool
         # Always has at least its name
         return len(self) > 1
     __bool__ = __nonzero__
 
     def __getattr__(self, attr):
+        # type: (str) -> _K
         try:
-            return object.__getattribute__(self, attr)
+            return object.__getattribute__(self, attr)  # type: ignore
         except AttributeError:
-            for k, v in six.iteritems(self.__dict__):
+            for k, v in self.d.items():
                 if self.ident(v) == attr:
                     return k
+        raise AttributeError
 
     def __dir__(self):
+        # type: () -> List[str]
         return [self.ident(x) for x in self.itervalues()]
+
+    def __reduce__(self):
+        # type: () -> Tuple[Type[Self], Tuple[str], Tuple[Dict[_K, _V]]]
+        return (self.__class__, (self._name,), (self.d,))
+
+    def __setstate__(self, state):
+        # type: (Tuple[Dict[_K, _V]]) -> Self
+        self.d.update(state[0])
+        return self
