@@ -1,17 +1,8 @@
-#############################################################################
-#  ipsec.py --- IPsec support for Scapy                                     #
-#                                                                           #
-#  Copyright (C) 2014  6WIND                                                #
-#                                                                           #
-#  This program is free software; you can redistribute it and/or modify it  #
-#  under the terms of the GNU General Public License version 2 as           #
-#  published by the Free Software Foundation.                               #
-#                                                                           #
-#  This program is distributed in the hope that it will be useful, but      #
-#  WITHOUT ANY WARRANTY; without even the implied warranty of               #
-#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU        #
-#  General Public License for more details.                                 #
-#############################################################################
+# SPDX-License-Identifier: GPL-2.0-only
+# This file is part of Scapy
+# See https://scapy.net/ for more information
+# Copyright (C) 2014 6WIND
+
 r"""
 IPsec layer
 ===========
@@ -19,7 +10,7 @@ IPsec layer
 Example of use:
 
 >>> sa = SecurityAssociation(ESP, spi=0xdeadbeef, crypt_algo='AES-CBC',
-...                          crypt_key='sixteenbytes key')
+...                          crypt_key=b'sixteenbytes key')
 >>> p = IP(src='1.1.1.1', dst='2.2.2.2')
 >>> p /= TCP(sport=45012, dport=80)
 >>> p /= Raw('testdata')
@@ -39,7 +30,6 @@ Example of use:
 True
 """
 
-from __future__ import absolute_import
 try:
     from math import gcd
 except ImportError:
@@ -47,17 +37,32 @@ except ImportError:
 import os
 import socket
 import struct
+import warnings
 
 from scapy.config import conf, crypto_validator
 from scapy.compat import orb, raw
 from scapy.data import IP_PROTOS
 from scapy.error import log_loading
-from scapy.fields import ByteEnumField, ByteField, IntField, PacketField, \
-    ShortField, StrField, XIntField, XStrField, XStrLenField
-from scapy.packet import Packet, bind_layers, Raw
+from scapy.fields import (
+    ByteEnumField,
+    ByteField,
+    IntField,
+    PacketField,
+    ShortField,
+    StrField,
+    XByteField,
+    XIntField,
+    XStrField,
+    XStrLenField,
+)
+from scapy.packet import (
+    Packet,
+    Raw,
+    bind_bottom_up,
+    bind_layers,
+    bind_top_down,
+)
 from scapy.layers.inet import IP, UDP
-import scapy.modules.six as six
-from scapy.modules.six.moves import range
 from scapy.layers.inet6 import IPv6, IPv6ExtHdrHopByHop, IPv6ExtHdrDestOpt, \
     IPv6ExtHdrRouting
 
@@ -87,7 +92,7 @@ class AH(Packet):
         ByteEnumField('nh', None, IP_PROTOS),
         ByteField('payloadlen', None),
         ShortField('reserved', None),
-        XIntField('spi', 0x0),
+        XIntField('spi', 0x00000001),
         IntField('seq', 0),
         XStrLenField('icv', None, length_from=__get_icv_len),
         # Padding len can only be known with the SecurityAssociation.auth_algo
@@ -120,11 +125,22 @@ class ESP(Packet):
     name = 'ESP'
 
     fields_desc = [
-        XIntField('spi', 0x0),
+        XIntField('spi', 0x00000001),
         IntField('seq', 0),
         XStrField('data', None),
     ]
 
+    @classmethod
+    def dispatch_hook(cls, _pkt=None, *args, **kargs):
+        if _pkt:
+            if len(_pkt) >= 4 and struct.unpack("!I", _pkt[0:4])[0] == 0x00:
+                return NON_ESP
+            elif len(_pkt) == 1 and struct.unpack("!B", _pkt)[0] == 0xff:
+                return NAT_KEEPALIVE
+            else:
+                return ESP
+        return cls
+
     overload_fields = {
         IP: {'proto': socket.IPPROTO_ESP},
         IPv6: {'nh': socket.IPPROTO_ESP},
@@ -134,10 +150,27 @@ class ESP(Packet):
     }
 
 
+class NON_ESP(Packet):  # RFC 3948, section 2.2
+    fields_desc = [
+        XIntField("non_esp", 0x0)
+    ]
+
+
+class NAT_KEEPALIVE(Packet):  # RFC 3948, section 2.2
+    fields_desc = [
+        XByteField("nat_keepalive", 0xFF)
+    ]
+
+
 bind_layers(IP, ESP, proto=socket.IPPROTO_ESP)
 bind_layers(IPv6, ESP, nh=socket.IPPROTO_ESP)
-bind_layers(UDP, ESP, dport=4500)  # NAT-Traversal encapsulation
-bind_layers(UDP, ESP, sport=4500)  # NAT-Traversal encapsulation
+
+# NAT-Traversal encapsulation
+bind_bottom_up(UDP, ESP, dport=4500)
+bind_bottom_up(UDP, ESP, sport=4500)
+bind_top_down(UDP, ESP, dport=4500, sport=4500)
+bind_top_down(UDP, NON_ESP, dport=4500, sport=4500)
+bind_top_down(UDP, NAT_KEEPALIVE, dport=4500, sport=4500)
 
 ###############################################################################
 
@@ -170,10 +203,18 @@ if conf.crypto_valid:
     from cryptography.exceptions import InvalidTag
     from cryptography.hazmat.backends import default_backend
     from cryptography.hazmat.primitives.ciphers import (
+        aead,
         Cipher,
         algorithms,
         modes,
     )
+    try:
+        # cryptography > 43.0
+        from cryptography.hazmat.decrepit.ciphers import (
+            algorithms as decrepit_algorithms
+        )
+    except ImportError:
+        decrepit_algorithms = algorithms
 else:
     log_loading.info("Can't import python-cryptography v1.7+. "
                      "Disabled IPsec encryption/authentication.")
@@ -225,11 +266,18 @@ class CryptAlgo(object):
         self.mode = mode
         self.icv_size = icv_size
 
-        if modes and self.mode is not None:
-            self.is_aead = issubclass(self.mode,
-                                      modes.ModeWithAuthenticationTag)
-        else:
-            self.is_aead = False
+        self.is_aead = False
+        # If using cryptography.hazmat.primitives.cipher.aead
+        self.ciphers_aead_api = False
+
+        if modes:
+            if self.mode is not None:
+                self.is_aead = issubclass(self.mode,
+                                          modes.ModeWithAuthenticationTag)
+            elif self.cipher in (aead.AESGCM, aead.AESCCM,
+                                 aead.ChaCha20Poly1305):
+                self.is_aead = True
+                self.ciphers_aead_api = True
 
         if block_size is not None:
             self.block_size = block_size
@@ -339,36 +387,53 @@ class CryptAlgo(object):
 
         return esp
 
-    def encrypt(self, sa, esp, key, esn_en=False, esn=0):
+    def encrypt(self, sa, esp, key, icv_size=None, esn_en=False, esn=0):
         """
         Encrypt an ESP packet
 
         :param sa:   the SecurityAssociation associated with the ESP packet.
         :param esp:  an unencrypted _ESPPlain packet with valid padding
         :param key:  the secret key used for encryption
+        :param icv_size: the length of the icv used for integrity check
         :esn_en:     extended sequence number enable which allows to use 64-bit
                      sequence number instead of 32-bit when using an AEAD
                      algorithm
         :esn:        extended sequence number (32 MSB)
         :return:    a valid ESP packet encrypted with this algorithm
         """
+        if icv_size is None:
+            icv_size = self.icv_size if self.is_aead else 0
         data = esp.data_for_encryption()
 
         if self.cipher:
             mode_iv = self._format_mode_iv(algo=self, sa=sa, iv=esp.iv)
-            cipher = self.new_cipher(key, mode_iv)
-            encryptor = cipher.encryptor()
-
+            aad = None
             if self.is_aead:
                 if esn_en:
                     aad = struct.pack('!LLL', esp.spi, esn, esp.seq)
                 else:
                     aad = struct.pack('!LL', esp.spi, esp.seq)
-                encryptor.authenticate_additional_data(aad)
-                data = encryptor.update(data) + encryptor.finalize()
-                data += encryptor.tag[:self.icv_size]
+            if self.ciphers_aead_api:
+                # New API
+                if self.cipher == aead.AESCCM:
+                    cipher = self.cipher(key, tag_length=icv_size)
+                else:
+                    cipher = self.cipher(key)
+                if self.name == 'AES-NULL-GMAC':
+                    # Special case for GMAC (rfc 4543 sect 3)
+                    data = data + cipher.encrypt(mode_iv, b"", aad + esp.iv + data)
+                else:
+                    data = cipher.encrypt(mode_iv, data, aad)
             else:
-                data = encryptor.update(data) + encryptor.finalize()
+                cipher = self.new_cipher(key, mode_iv)
+                encryptor = cipher.encryptor()
+
+                if self.is_aead:
+                    encryptor.authenticate_additional_data(aad)
+                    data = encryptor.update(data) + encryptor.finalize()
+                    data += encryptor.tag[:icv_size]
+                else:
+                    data = encryptor.update(data) + encryptor.finalize()
 
         return ESP(spi=esp.spi, seq=esp.seq, data=esp.iv + data)
 
@@ -397,29 +462,45 @@ class CryptAlgo(object):
 
         if self.cipher:
             mode_iv = self._format_mode_iv(sa=sa, iv=iv)
-            cipher = self.new_cipher(key, mode_iv, icv)
-            decryptor = cipher.decryptor()
-
+            aad = None
             if self.is_aead:
-                # Tag value check is done during the finalize method
                 if esn_en:
-                    decryptor.authenticate_additional_data(
-                        struct.pack('!LLL', esp.spi, esn, esp.seq))
+                    aad = struct.pack('!LLL', esp.spi, esn, esp.seq)
                 else:
-                    decryptor.authenticate_additional_data(
-                        struct.pack('!LL', esp.spi, esp.seq))
-            try:
-                data = decryptor.update(data) + decryptor.finalize()
-            except InvalidTag as err:
-                raise IPSecIntegrityError(err)
+                    aad = struct.pack('!LL', esp.spi, esp.seq)
+            if self.ciphers_aead_api:
+                # New API
+                if self.cipher == aead.AESCCM:
+                    cipher = self.cipher(key, tag_length=icv_size)
+                else:
+                    cipher = self.cipher(key)
+                try:
+                    if self.name == 'AES-NULL-GMAC':
+                        # Special case for GMAC (rfc 4543 sect 3)
+                        data = data + cipher.decrypt(mode_iv, icv, aad + iv + data)
+                    else:
+                        data = cipher.decrypt(mode_iv, data + icv, aad)
+                except InvalidTag as err:
+                    raise IPSecIntegrityError(err)
+            else:
+                cipher = self.new_cipher(key, mode_iv, icv)
+                decryptor = cipher.decryptor()
+
+                if self.is_aead:
+                    # Tag value check is done during the finalize method
+                    decryptor.authenticate_additional_data(aad)
+                try:
+                    data = decryptor.update(data) + decryptor.finalize()
+                except InvalidTag as err:
+                    raise IPSecIntegrityError(err)
 
         # extract padlen and nh
         padlen = orb(data[-2])
         nh = orb(data[-1])
 
         # then use padlen to determine data and padding
-        data = data[:len(data) - padlen - 2]
         padding = data[len(data) - padlen - 2: len(data) - 2]
+        data = data[:len(data) - padlen - 2]
 
         return _ESPPlain(spi=esp.spi,
                          seq=esp.seq,
@@ -447,36 +528,80 @@ if algorithms:
     CRYPT_ALGOS['AES-CTR'] = CryptAlgo('AES-CTR',
                                        cipher=algorithms.AES,
                                        mode=modes.CTR,
+                                       block_size=1,
                                        iv_size=8,
                                        salt_size=4,
                                        format_mode_iv=_aes_ctr_format_mode_iv)
     _salt_format_mode_iv = lambda sa, iv, **kw: sa.crypt_salt + iv
     CRYPT_ALGOS['AES-GCM'] = CryptAlgo('AES-GCM',
-                                       cipher=algorithms.AES,
-                                       mode=modes.GCM,
+                                       cipher=aead.AESGCM,
+                                       key_size=(16, 24, 32),
+                                       mode=None,
                                        salt_size=4,
+                                       block_size=1,
                                        iv_size=8,
                                        icv_size=16,
                                        format_mode_iv=_salt_format_mode_iv)
-    if hasattr(modes, 'CCM'):
-        CRYPT_ALGOS['AES-CCM'] = CryptAlgo('AES-CCM',
-                                           cipher=algorithms.AES,
-                                           mode=modes.CCM,
-                                           iv_size=8,
-                                           salt_size=3,
-                                           icv_size=16,
-                                           format_mode_iv=_salt_format_mode_iv)
-    # XXX: RFC7321 states that DES *MUST NOT* be implemented.
-    # XXX: Keep for backward compatibility?
+    # GMAC: rfc 4543, "companion to the AES Galois/Counter Mode ESP"
+    # This is defined as a crypt_algo by rfc, but has the role of an auth_algo
+    CRYPT_ALGOS['AES-NULL-GMAC'] = CryptAlgo('AES-NULL-GMAC',
+                                             cipher=aead.AESGCM,
+                                             key_size=(16, 24, 32),
+                                             mode=None,
+                                             salt_size=4,
+                                             block_size=1,
+                                             iv_size=8,
+                                             icv_size=16,
+                                             format_mode_iv=_salt_format_mode_iv)
+    CRYPT_ALGOS['AES-CCM'] = CryptAlgo('AES-CCM',
+                                       cipher=aead.AESCCM,
+                                       mode=None,
+                                       key_size=(16, 24, 32),
+                                       block_size=1,
+                                       iv_size=8,
+                                       salt_size=3,
+                                       icv_size=16,
+                                       format_mode_iv=_salt_format_mode_iv)
+    CRYPT_ALGOS['CHACHA20-POLY1305'] = CryptAlgo('CHACHA20-POLY1305',
+                                                 cipher=aead.ChaCha20Poly1305,
+                                                 mode=None,
+                                                 key_size=32,
+                                                 block_size=1,
+                                                 iv_size=8,
+                                                 salt_size=4,
+                                                 icv_size=16,
+                                                 format_mode_iv=_salt_format_mode_iv)  # noqa: E501
+
     # Using a TripleDES cipher algorithm for DES is done by using the same 64
     # bits key 3 times (done by cryptography when given a 64 bits key)
     CRYPT_ALGOS['DES'] = CryptAlgo('DES',
-                                   cipher=algorithms.TripleDES,
+                                   cipher=decrepit_algorithms.TripleDES,
                                    mode=modes.CBC,
                                    key_size=(8,))
     CRYPT_ALGOS['3DES'] = CryptAlgo('3DES',
-                                    cipher=algorithms.TripleDES,
+                                    cipher=decrepit_algorithms.TripleDES,
                                     mode=modes.CBC)
+    if decrepit_algorithms is algorithms:
+        # cryptography < 43 raises a DeprecationWarning
+        from cryptography.utils import CryptographyDeprecationWarning
+        with warnings.catch_warnings():
+            # Hide deprecation warnings
+            warnings.filterwarnings("ignore",
+                                    category=CryptographyDeprecationWarning)
+            CRYPT_ALGOS['CAST'] = CryptAlgo('CAST',
+                                            cipher=decrepit_algorithms.CAST5,
+                                            mode=modes.CBC)
+            CRYPT_ALGOS['Blowfish'] = CryptAlgo('Blowfish',
+                                                cipher=decrepit_algorithms.Blowfish,
+                                                mode=modes.CBC)
+    else:
+        CRYPT_ALGOS['CAST'] = CryptAlgo('CAST',
+                                        cipher=decrepit_algorithms.CAST5,
+                                        mode=modes.CBC)
+        CRYPT_ALGOS['Blowfish'] = CryptAlgo('Blowfish',
+                                            cipher=decrepit_algorithms.Blowfish,
+                                            mode=modes.CBC)
+
 
 ###############################################################################
 if conf.crypto_valid:
@@ -557,16 +682,17 @@ class AuthAlgo(object):
         mac = self.new_mac(key)
 
         if pkt.haslayer(ESP):
-            mac.update(raw(pkt[ESP]))
+            mac.update(bytes(pkt[ESP]))
+            if esn_en:
+                # RFC4303 sect 2.2.1
+                mac.update(struct.pack('!L', esn))
             pkt[ESP].data += mac.finalize()[:self.icv_size]
 
         elif pkt.haslayer(AH):
-            clone = zero_mutable_fields(pkt.copy(), sending=True)
+            mac.update(bytes(zero_mutable_fields(pkt.copy(), sending=True)))
             if esn_en:
-                temp = raw(clone) + struct.pack('!L', esn)
-            else:
-                temp = raw(clone)
-            mac.update(temp)
+                # RFC4302 sect 2.5.1
+                mac.update(struct.pack('!L', esn))
             pkt[AH].icv = mac.finalize()[:self.icv_size]
 
         return pkt
@@ -595,7 +721,10 @@ class AuthAlgo(object):
             pkt_icv = pkt.data[len(pkt.data) - self.icv_size:]
             clone = pkt.copy()
             clone.data = clone.data[:len(clone.data) - self.icv_size]
-            temp = raw(clone)
+            mac.update(bytes(clone))
+            if esn_en:
+                # RFC4303 sect 2.2.1
+                mac.update(struct.pack('!L', esn))
 
         elif pkt.haslayer(AH):
             if len(pkt[AH].icv) != self.icv_size:
@@ -604,12 +733,11 @@ class AuthAlgo(object):
                 pkt[AH].icv = pkt[AH].icv[:self.icv_size]
             pkt_icv = pkt[AH].icv
             clone = zero_mutable_fields(pkt.copy(), sending=False)
+            mac.update(bytes(clone))
             if esn_en:
-                temp = raw(clone) + struct.pack('!L', esn)
-            else:
-                temp = raw(clone)
+                # RFC4302 sect 2.5.1
+                mac.update(struct.pack('!L', esn))
 
-        mac.update(temp)
         computed_icv = mac.finalize()[:self.icv_size]
 
         # XXX: Cannot use mac.verify because the ICV can be truncated
@@ -801,7 +929,9 @@ class SecurityAssociation(object):
     SUPPORTED_PROTOS = (IP, IPv6)
 
     def __init__(self, proto, spi, seq_num=1, crypt_algo=None, crypt_key=None,
-                 auth_algo=None, auth_key=None, tunnel_header=None, nat_t_header=None, esn_en=False, esn=0):   # noqa: E501
+                 crypt_icv_size=None,
+                 auth_algo=None, auth_key=None,
+                 tunnel_header=None, nat_t_header=None, esn_en=False, esn=0):
         """
         :param proto: the IPsec proto to use (ESP or AH)
         :param spi: the Security Parameters Index of this SA
@@ -809,6 +939,8 @@ class SecurityAssociation(object):
                         packets
         :param crypt_algo: the encryption algorithm name (only used with ESP)
         :param crypt_key: the encryption key (only used with ESP)
+        :param crypt_icv_size: change the default size of the crypt_algo
+                               (only used with ESP)
         :param auth_algo: the integrity algorithm name
         :param auth_key: the integrity key
         :param tunnel_header: an instance of a IP(v6) header that will be used
@@ -821,10 +953,10 @@ class SecurityAssociation(object):
         :param esn: extended sequence number (32 MSB)
         """
 
-        if proto not in (ESP, AH, ESP.name, AH.name):
+        if proto not in {ESP, AH, ESP.name, AH.name}:
             raise ValueError("proto must be either ESP or AH")
-        if isinstance(proto, six.string_types):
-            self.proto = eval(proto)
+        if isinstance(proto, str):
+            self.proto = {ESP.name: ESP, AH.name: AH}[proto]
         else:
             self.proto = proto
 
@@ -851,6 +983,7 @@ class SecurityAssociation(object):
             self.crypt_algo = CRYPT_ALGOS['NULL']
             self.crypt_key = None
             self.crypt_salt = None
+        self.crypt_icv_size = crypt_icv_size
 
         if auth_algo:
             if auth_algo not in AUTH_ALGOS:
@@ -907,10 +1040,14 @@ class SecurityAssociation(object):
 
         esp = self.crypt_algo.pad(esp)
         esp = self.crypt_algo.encrypt(self, esp, self.crypt_key,
+                                      self.crypt_icv_size,
                                       esn_en=esn_en or self.esn_en,
                                       esn=esn or self.esn)
 
-        self.auth_algo.sign(esp, self.auth_key)
+        self.auth_algo.sign(esp,
+                            self.auth_key,
+                            esn_en=esn_en or self.esn_en,
+                            esn=esn or self.esn)
 
         if self.nat_t_header:
             nat_t_header = self.nat_t_header.copy()
@@ -923,17 +1060,16 @@ class SecurityAssociation(object):
             ip_header /= nat_t_header
 
         if ip_header.version == 4:
-            ip_header.len = len(ip_header) + len(esp)
+            del ip_header.len
             del ip_header.chksum
-            ip_header = ip_header.__class__(raw(ip_header))
         else:
-            ip_header.plen = len(ip_header.payload) + len(esp)
+            del ip_header.plen
 
         # sequence number must always change, unless specified by the user
         if seq_num is None:
             self.seq_num += 1
 
-        return ip_header / esp
+        return ip_header.__class__(raw(ip_header / esp))
 
     def _encrypt_ah(self, pkt, seq_num=None, esn_en=False, esn=0):
 
@@ -1022,9 +1158,12 @@ class SecurityAssociation(object):
 
         if verify:
             self.check_spi(pkt)
-            self.auth_algo.verify(encrypted, self.auth_key)
+            self.auth_algo.verify(encrypted, self.auth_key,
+                                  esn_en=esn_en or self.esn_en,
+                                  esn=esn or self.esn)
 
         esp = self.crypt_algo.decrypt(self, encrypted, self.crypt_key,
+                                      self.crypt_icv_size or
                                       self.crypt_algo.icv_size or
                                       self.auth_algo.icv_size,
                                       esn_en=esn_en or self.esn_en,
@@ -1052,8 +1191,13 @@ class SecurityAssociation(object):
                 # recompute checksum
                 ip_header = ip_header.__class__(raw(ip_header))
             else:
-                encrypted.underlayer.nh = esp.nh
-                encrypted.underlayer.remove_payload()
+                if self.nat_t_header:
+                    # drop the UDP header and return the payload untouched
+                    ip_header.nh = esp.nh
+                    ip_header.remove_payload()
+                else:
+                    encrypted.underlayer.nh = esp.nh
+                    encrypted.underlayer.remove_payload()
                 ip_header.plen = len(ip_header.payload) + len(esp.data)
 
             cls = ip_header.guess_payload_class(esp.data)
